from ..llm.llm_async_builder import LLMAsync
from ...util.db.postgres_vector_async import PostgresVectorAsync
from ...util.db.schema import EmbbedingSchame, DatasetSchema, KB_TABLE_NAME, ANSWER_TABLE_NAME, QUESTION_TABLE_NAME
from ...util.id_generator import generate_common_id
from ...util.logger import logger
import json
from scipy.spatial.distance import cosine

class DatasetBuilderAsync:     

    def __init__(self, db_driver: PostgresVectorAsync):
        """初始化知识库插件构建器。"""
        super().__init__()
        self.db_driver = db_driver
        self._connection = None
        self.answer_schema = EmbbedingSchame(ANSWER_TABLE_NAME)
        self.question_schema = EmbbedingSchame(QUESTION_TABLE_NAME)
        self.dataset_schema = DatasetSchema()

    async def create_dataset(self, name, creator="gpt_builder"):
        """向知识库添加一个新的条目，包括创建者和创建时间。"""
        async with (await self.db_driver.pool).acquire() as connection:
            async with connection.transaction():
                sql = f"""
                INSERT INTO {KB_TABLE_NAME} 
                ({self.dataset_schema.name}, {self.dataset_schema.creator}, {self.dataset_schema.created_at}) 
                VALUES ($1, $2, CURRENT_TIMESTAMP) RETURNING {self.dataset_schema.id}
                """
                logger.info(f"Inserting KB entry: {name} by {creator} sql {sql}")
                new_id = await connection.fetchval(sql, name, creator)
            return new_id

    async def get_dataset(self, filters, creator="gpt_builder"):
        """
        异步根据指定的过滤条件获取知识库条目。
        
        参数:
            creator (str, optional): 创建者的名称。默认为 "gpt_builder"。
            filters (dict): 过滤条件的键值对。
        
        返回:
            list: 符合条件的知识库条目列表，每个条目是一个字典。
        """
        # 构建过滤条件的 SQL 语句
        filter_conditions = " AND ".join([f"{key} = ${i+1}" for i, key in enumerate(filters.keys())])
        filter_values = list(filters.values())
        
        # 添加 creator 过滤条件
        filter_conditions += f" AND {self.dataset_schema.creator} = ${len(filters) + 1}"
        filter_values.append(creator)
        
        query = f"""
        SELECT * FROM {KB_TABLE_NAME}
        WHERE {filter_conditions}
        """
        
        async with (await self.db_driver.pool).acquire() as connection:
            async with connection.transaction():
                results = await connection.fetch(query, *filter_values)
        
        return [dict(result) for result in results]

    async def create_datas(self, id, index_text, questions_list):
        """向数据库添加一个答案和相关的问题列表。"""
        if not id:
            logger.info("未成功生产知识库实例")
            return
        async with (await self.db_driver.pool).acquire() as connection:
            async with connection.transaction():
                try:
                    # 插入索引并获取ID
                    sql_query = self.db_driver.get_insert_into_str(ANSWER_TABLE_NAME, [self.answer_schema.text, self.answer_schema.vector, self.answer_schema.kb_id]) + f" RETURNING {self.answer_schema.id}"
                    logger.info(f"sql_query {sql_query}")
                    index_vector = await self.generate_vector(index_text)
                    if not index_vector:
                        logger.info(f"{index_text} 未成功向量化")
                        return
                    index_vector = json.dumps(index_vector)
                    index_id = await connection.fetchval(
                        sql_query,
                        index_text, index_vector, id
                    )
                    logger.info(f"index_id {index_id}")
                    # 批量插入问题
                    question_entries = []
                    for question_text in questions_list:
                        question_vector = await self.generate_vector(question_text)
                        if question_vector:
                            question_vector = json.dumps(question_vector)
                            question_entries.append((index_id, question_text, question_vector))
                        else:
                            logger.info(f"{question_text} 未成功向量化")
                        
                    if question_entries:
                        sql_query = self.db_driver.get_insert_into_str(QUESTION_TABLE_NAME, [self.answer_schema.id, self.question_schema.text, self.question_schema.vector]) + f" RETURNING {self.question_schema.id}"
                        await connection.executemany(sql_query,question_entries)
                except Exception as e:
                    logger.error(f"数据库操作失败: {e}")
                    # 事务回滚是自动的，但仍可以记录或处理异常
                    raise


    async def generate_vector(self, text):
        """生成文本的向量表示，使用spaCy库。

        Args:
            text (str): 输入文本。

        Returns:
            blob: 向量的BLOB表示。
        """
        response = await LLMAsync.embedding(text)
        if not response:
            return {}
        embedding = response.get("data",{}).get("data", [])[0] if response.get("data") else {}
        if not embedding and embedding.get("embedding"):
            return
        return embedding.get("embedding")

    async def query_similarity(self, text, dataset_ids=[], similarity=-1, search_all=False):
        """
        使用联合查询来计算输入文本与数据库中内容的余弦相似度，并返回相关信息。

        Args:
            dataset_ids (list): 知识库ID列表。
            text (str): 查询的文本。
            similarity (float): 用于过滤的相似度阈值。
            search_all (bool): 是否计算加权平均相似度。

        Returns:
            dict: 以知识库ID为键，每个键包含相关内容的列表。
        """
        if not dataset_ids:
            logger.info("未成功获取知识库实例")
            return None

        input_vector = await self.generate_vector(text)
        if not input_vector:
            logger.info("输入文本未成功向量化")
            return None

        kb_content_map = {}
        async with (await self.db_driver.pool).acquire() as connection:
            async with connection.transaction():
                for dataset_id in dataset_ids:
                    sql = f"""
                        SELECT a.{self.answer_schema.vector}, a.{self.answer_schema.text}, a.{self.answer_schema.id},
                            q.{self.question_schema.vector}, q.{self.question_schema.text}
                        FROM {ANSWER_TABLE_NAME} AS a
                        JOIN {QUESTION_TABLE_NAME} AS q ON a.{self.answer_schema.id} = q.{self.answer_schema.id}
                        WHERE a.{self.answer_schema.kb_id} = $1"""
                    logger.info(f"[gpt_builder] SQL:{sql}")
                    data_list = await connection.fetch(sql, dataset_id)

                    content_list = []
                    for answer_vector, answer_text, answer_id, question_vector, question_text in data_list:
                        answer_vector_data = json.loads(answer_vector)
                        question_vector_data = json.loads(question_vector)
                        answer_similarity = 1 - cosine(input_vector, answer_vector_data)
                        question_similarity = 1 - cosine(input_vector, question_vector_data)
                        total_similarity = (answer_similarity + question_similarity) / 2 if search_all else question_similarity

                        if total_similarity < similarity:
                            continue

                        content_list.append({
                            "index_id": answer_id,
                            "related_questions": question_text,
                            "answer_text": answer_text,
                            "similarity": total_similarity,
                            "search_all": search_all
                        })

                    if content_list:
                        content_list.sort(key=lambda x: x['similarity'], reverse=True)
                        kb_content_map[dataset_id] = content_list
        return kb_content_map
    
    async def query_regex(self, regex, dataset_ids=[]):
        """
        根据正则表达式查询数据库中符合条件的答案索引。

        Args:
            regex (str): 查询的正则表达式。
            dataset_ids (list): 知识库ID列表。

        Returns:
            dict: 以知识库ID为键，每个键包含匹配的索引ID和对应的索引文本及相关内容列表的字典。
        """

        if not dataset_ids:
            logger.info("未成功获取知识库实例")
            return None

        # 最外层字典，以kb_id为键
        kb_content_map = {}
        async with (await self.db_driver.pool).acquire() as connection:
            async with connection.transaction():
                for kb_id in dataset_ids:
                    # 使用连表查询每个kb_id对应的答案和问题数据
                    sql = f"""
                        SELECT a.{self.answer_schema.id} AS answer_id, a.{self.answer_schema.text} AS answer_text,
                            q.{self.question_schema.id} AS question_id, q.{self.question_schema.text} AS question_text
                        FROM {ANSWER_TABLE_NAME} AS a
                        JOIN {QUESTION_TABLE_NAME} AS q ON a.{self.answer_schema.id} = q.{self.answer_schema.id}
                        WHERE a.{self.answer_schema.kb_id} = %s AND q.{self.question_schema.text} ~ %s"""
                    data_list = await connection.fetch(sql, kb_id, regex)

                    for question_text, question_id, answer_id, answer_text in data_list:
                        if kb_id not in kb_content_map:
                            kb_content_map[kb_id] = []

                        # 添加到列表
                        kb_content_map[kb_id].append({
                            "answer_id": answer_id,
                            "related_answer": answer_text,
                            "question_id": question_id,
                            "questions_text": question_text
                        })

        return kb_content_map






from .util import logger
from .session_manager.session_manager import SessionManager
import json


class BaseBuilder:

    def __init__(self,  session_id, session_manager: SessionManager, template="Hello, {name}! How can I assist you today?"):
        self.template = template
        self.session_id = session_id
        self.session_manager = session_manager
        self._content = None

    def _set_template(self, new_template):
        """Set a new prompt template."""
        self.template = new_template
        return self
    
    def _render_template(self,  **kwargs):
        """Set a new prompt template."""
        try:
            self.content = self.template.format(**kwargs)
        except KeyError as e:
            logger.info(f"Missing a required key in the template: {e}")
            self._content = self.template  # Return the unmodified template if missing keys
        return self
    
    @staticmethod
    def generate_curl_command(url, json_data, headers):
        """生成curl命令的辅助方法"""
        curl_command = f"curl -X POST '{url}'"
        if headers:
            for key, value in headers.items():
                curl_command += f" -H '{key}: {value}'"
        if json_data:
            json_str = json.dumps(json_data)
            curl_command += f" -d '{json_str}'"
        return curl_command
    


    




# -*- coding: utf-8 -*-

import os


def current_path():
    return os.getcwd()

def create_dir_if_not_exists(path):
    if os.path.exists(path):
        return
    os.makedirs(path)

def join_path_filename(path, filename):
    return os.path.join(path, filename)



from .base_session import Session
from ..config.config_manager import config_manager
from ..util.logger import logger


"""
    e.g.  [
        {"role": "system", "content": "You are a helpful assistant."},
        {"role": "user", "content": "Who won the world series in 2020?"},
        {"role": "assistant", "content": "The Los Angeles Dodgers won the World Series in 2020."},
        {"role": "user", "content": "Where was it played?"}
    ]
"""


class ChatGPTSession(Session):

    def __init__(self, session_id, system_prompt, model):
        super().__init__(session_id, system_prompt)
        self.model = model
        self.reset()

    def discard_exceeding(self, cur_tokens=None):
        model_config = config_manager.get_model_config(self.model)
        
        if not model_config:
            logger.debug(f"Model configuration not found for: {self.model}")
            return 0

        # Get the token settings for the specified model
        max_tokens = model_config.get('max_tokens', 80000)
        precise = True
        try:
            cur_tokens = self.calc_tokens()
        except Exception as e:
            precise = False
            logger.debug("Exception when counting tokens precisely for query: {}".format(e))
            if cur_tokens is None:
                raise e
        while cur_tokens > max_tokens:
            if len(self.messages) > 2:
                self.messages.pop(1)
            elif len(self.messages) == 2 and self.messages[1]["role"] == "assistant":
                self.messages.pop(1)
                if precise:
                    cur_tokens = self.calc_tokens()
                else:
                    cur_tokens = cur_tokens - max_tokens
                break
            elif len(self.messages) == 2 and self.messages[1]["role"] == "user":
                logger.warning(f"user message exceed max_tokens. total_tokens={cur_tokens}")
                break
            else:
                logger.debug(f"max_tokens={max_tokens}, total_tokens={cur_tokens}, len(messages)={self.messages}")
                break
            if precise:
                cur_tokens = self.calc_tokens()
            else:
                cur_tokens = cur_tokens - max_tokens
        return cur_tokens

    def calc_tokens(self):
        return self.num_tokens_from_messages(self.messages, self.model)






# -*- coding: utf-8 -*-

"""用于函数执行时间自动计时并通过日志输出的decorator。示例:
  from common_sdk.system.function_timer import function_timer

  @function_timer(name='YourFunction')
  def func():
      xxx
"""

from .logger import logger
from .datetime_utils import DateTime


def function_timer(name=None):
    def decorator(func):
        def wrapper_function(*args, **kwargs):
            start = DateTime()
            result = func(*args,  **kwargs)
            milliseconds = DateTime().milliseconds - start.milliseconds
            func_name = name
            if func_name is None:
                func_name = f'{func.__name__!r}'
            logger.info('[{}]  函数执行时间: {}毫秒'.format(func_name, milliseconds))
            return result
        return wrapper_function

    return decorator


from ...prompt_base_builder import BaseBuilder
from ...prompt_template import KB_REFERENCE_TEMPLATE
from ...util.logger import logger
from ...session_manager import SessionManager

class PromptBuilder(BaseBuilder):

    @property
    def session(self):
        return self._session


    def __init__(self, session_id, session_manager: SessionManager):
        self.template = None
        self._session = None
        self.session_id = session_id
        self.session_manager = session_manager
        self._tasks = []
        self._content = None

    def set_system(self, system_template, **kwargs):
        """设置系统提示词

        Args:
            system_prompt (str): 提示词模板，
            kwargs： 提示词变量渲染参数
        """
        self._tasks.append(lambda: self._sync_set_system(system_template, **kwargs))
        return self

    def _sync_set_system(self, system_template, **kwargs):
        self._set_template(system_template)
        self._render_template(**kwargs)
        self._content = self.content
        self._session = self.session_manager.build_session(self.session_id, self._content)
        logger.info(f"set_system res {self._session.messages}")

    def set_prompt(self, prompt_template, **kwargs):
        self._tasks.append(lambda: self._sync_set_prompt(prompt_template, **kwargs))
        return self

    def _sync_set_prompt(self, prompt_template, **kwargs):
        self._set_template(prompt_template)
        self._render_template(**kwargs)
        self._content = self.content
        self._session = self.session_manager.session_query(self.session_id, self._content)
        logger.info(f"set_prompt res {self._session.messages}")

    def set_kb_reference(self, **kwargs):
        self._tasks.append(lambda: self._sync_set_kb_reference(**kwargs))
        return self

    def _sync_set_kb_reference(self, **kwargs):
        kb_reference_template = KB_REFERENCE_TEMPLATE
        self._set_template(kb_reference_template)
        self._render_template(**kwargs)
        self._content = self.content
        self._session = self.session_manager.session_query(self.session_id, self._content)
        logger.info(f"set_kb_reference res {self._session.messages}")

    def build(self):
        """执行所有任务"""
        for task in self._tasks:
            task()
        return self


    




BASE_URL = "https://www.lazygpt.cn/api"
API_KEY = ""
MODEL_SETTINGS = [
        {
            "model": "gpt-3.5-turbo",
            "max_tokens": 8000,
            "token_setting": "gpt-3.5-turbo"
        },
        {
            "model": "gpt-3.5-turbo-16k",
            "max_tokens": 16000,
            "token_setting": "gpt-4"
        },
        {
            "model": "gpt-4",
            "max_tokens": 8000,
            "token_setting": "gpt-4"
        },
        {
            "model": "gpt-4-0125-preview",
            "max_tokens": 128000,
            "token_setting": "gpt-4"
        },
        {
            "model": "gpt-4-vision-preview",
            "max_tokens": 128000,
            "token_setting": "gpt-4"
        }
    ]

TOKEN_SETTINGS = {
        "gpt-3.5-turbo": {
            "tokens_per_message": 4,
            "tokens_per_name": -1
        },
        "gpt-4": {
            "tokens_per_message": 3,
            "tokens_per_name": 1
        }
    }







from .chatgpt_session import ChatGPTSession
from .session_manager import SessionManager
from .session_manager_async import SessionManagerAsync
from .storage.global_storage import global_storage
from .storage.redis_storage import RedisStorage
from .storage.redis_storage_async import RedisStorageAsync


# -*- coding: utf-8 -*-

import uuid


def generate_common_id(with_hyphen=False, is_uppercase=False):
    id = str(uuid.uuid4())
    if not with_hyphen:
        id = id.replace('-', '')
    if is_uppercase:
        id = id.upper()
    return id
    
import hashlib




# -*- coding: utf-8 -*-

from contextvars import ContextVar

HTTP_MESSAGE_UUID = "message_id"
HTTP_USER_IP = "user_ip"
HTTP_REQUEST_TIMESTAMP = "request_timestamp"
COMMON_CONTEXT = 'context'

_message_id_ctx_var: ContextVar[str] = ContextVar(HTTP_MESSAGE_UUID, default=None)
_user_ip_ctx_var: ContextVar[str] = ContextVar(HTTP_USER_IP, default="0.0.0.0")
_request_timestamp_ctx_var: ContextVar[int] = ContextVar(HTTP_REQUEST_TIMESTAMP, default=None)
_common_ctx_var: ContextVar[dict] = ContextVar(COMMON_CONTEXT, default={})


def get(key: str):
    return _common_ctx_var.get().get(key)


def set(key: str, val) -> None:
    _common_ctx_var.get().update({key: val})


def get_message_uuid() -> str:
    return _message_id_ctx_var.get()


def set_message_uuid(id: str) -> str:
    return _message_id_ctx_var.set(id)

def get_user_ip() -> str:
    return _user_ip_ctx_var.get()


def set_user_ip(ip: str) -> str:
    return _user_ip_ctx_var.set(ip)

def get_request_timestamp() -> int:
    return _request_timestamp_ctx_var.get()


def set_request_timestamp(timestamp: int) -> int:
    return _request_timestamp_ctx_var.set(timestamp)


# chatGPT的session管理器,用来存放进程中的一些全局变量，比如上下文回话
from ..util.logger import logger
from .chatgpt_session import ChatGPTSession

from .storage.redis_storage_async import RedisStorageAsync
from typing import Type, Optional


class SessionManagerAsync(object):
    """这里主要是更新真实的上下文缓存"""

    def __init__(self, sessioncls: Type[ChatGPTSession], session_storage: RedisStorageAsync, model):
        self.session_storage = session_storage
        self.sessioncls = sessioncls
        self.expires_in_seconds = 3600
        self.model = model

    async def build_session(self, session_id, system_prompt=None):
        if session_id is None:
            return self.sessioncls(session_id)
        
        session_json = await self.session_storage.get(session_id)
        if session_json is None:
            session = self.sessioncls(session_id, system_prompt, self.model)
            await self.session_storage.set(session_id, session.to_json())
        else:
            session = self.sessioncls.from_json(session_json, self.model)
            if system_prompt is not None:
                self.sessioncls.set_system_prompt(system_prompt)
            await self.session_storage.set(session_id, session.to_json(), self.expires_in_seconds)
        
        return session

    async def session_query(self, session_id, query):
        session = await self.build_session(session_id)
        session.add_query(query)
        await self.session_storage.set(session_id, session.to_json(), self.expires_in_seconds)
        try:
            total_tokens = session.discard_exceeding()
            logger.debug("prompt tokens used={}".format(total_tokens))
        except Exception as e:
            logger.debug("Exception when counting tokens precisely for prompt: {}".format(str(e)))
        return session

    async def session_reply(self, session_id, reply, total_tokens=None):
        session = await self.build_session(session_id)
        session.add_reply(reply)
        await self.session_storage.set(session_id, session.to_json(), self.expires_in_seconds)    
        try:
            tokens_cnt = session.discard_exceeding(total_tokens)
            logger.debug("raw total_tokens={}, savesession tokens={}".format(total_tokens, tokens_cnt))
        except Exception as e:
            logger.debug("Exception when counting tokens precisely for session: {}".format(str(e)))
        return session

    async def clear_session(self, session_id):
        await self.session_storage.delete(session_id)



# -*- coding: utf-8 -*-
from ...util.singleton import SingletonMetaThreadSafe
import redis
import json


class RedisStorage(metaclass=SingletonMetaThreadSafe):

    def __init__(self, url) -> None:
        self._redis = redis.Redis(url=url)

    def set(self, key, data, expired=7200) -> bool:
        if not key:
            raise Exception("找不到key")
        return self._redis.setex(key, expired, json.dumps(data))

    def get(self, key) -> dict:
        if not key:
            raise Exception("找不到key")
        data = self._redis.get(key)
        if not data:
            return {}
        return json.loads(data)

    def check(self, key, expired=24 * 60 * 60) -> bool:
        if not key:
            raise Exception("找不到key")
        return self._redis.expire(key, expired)

    def delete(self, key) -> None:
        if not key:
            raise Exception("找不到key")
        self._redis.delete(key)
    
    def enqueue_message(self, message) -> None:
        """将消息添加到队列中"""
        self._redis.lpush(self.COMMENT_QUEUE_KEY, json.dumps(message))
    
    def dequeue_message(self, timeout=0) -> dict:
        """从队列中移除并返回一条消息"""
        message = self._redis.brpop(self.COMMENT_QUEUE_KEY, timeout=timeout)
        if message:
            return json.loads(message[1])
        return None

    def acquire_lock(self, lock_name, lock_timeout=60):
        """尝试获取锁，成功则返回 True，否则返回 False。"""
        result = self._redis.set(lock_name, 1, ex=lock_timeout, nx=True)
        return result is not None

    def release_lock(self, lock_name):
        """释放锁。"""
        self._redis.delete(lock_name)




# -*- coding: utf-8 -*-

import logging
from logging.handlers import SysLogHandler
from uuid import uuid1

from .singleton import SingletonMetaThreadSafe as SingletonMetaclass
from . import sys_env,file_utils, context

APPNAME_ENV_NAME = 'LazyAgent'
# logger
LOGGER_CATEGORY_ENV_NAME = "INFO,DEBUG,ERROR"
# console
LOGGER_ENABLE_CONSOLE_ENV_NAME = True
# syslog
LOGGER_ENABLE_SYSLOG_ENV_NAME = False
LOGGER_SYSLOG_HOST_ENV_NAME = "logger.server"
LOGGER_SYSLOG_PORT_ENV_NAME = 514
LOGGER_SYSLOG_FACILITY_ENV_NAME = "local7"
# FILE
LOGGER_ENABLE_FILE_ENV_NAME = False
LOGGER_FILE_DIRECTORY_ENV_NAME = "$HOME/logs/$APPNAME/$VERSION"


class Logger(metaclass=SingletonMetaclass):

    def __init__(self):
        self._formatter = None
        self._logger = logging.getLogger()
        self._logger.setLevel(logging.INFO)
        self.__init_syslog_handler()
        self.__init_console_handler()
        self.__init_file_handler()
        self.logger.info(self.__wrap_message_with_uuid(f"########日志类初始化#######"))

    @property
    def name(self):
        name = sys_env.get_env(APPNAME_ENV_NAME, default="未知应用名称")
        return name

    @property
    def formatter(self):
        if self._formatter is not None:
            return self._formatter
        formatter = f'{self.name} | %(asctime)s | %(levelname)s | %(name)s | pid:%(process)d@%(pathname)s:%(lineno)s | %(message)s'
        self._formatter = logging.Formatter(formatter)
        return self._formatter

    @property
    def message_uuid(self):
        try:
            message_uuid = context.get_message_uuid()
        except Exception:
            pass
        if not message_uuid:
            message_uuid = f"{uuid1()}".replace('-', '')
            context.set_message_uuid(message_uuid)
        return message_uuid

    @property
    def logger(self):
        return self._logger

    def exc_info(self):
        return sys_env.get_env('LOGGER_EXC_INFO', False)

    def debug(self, message, *args, **kwargs):
        message = self.__wrap_message_with_uuid(message)
        kwargs = dict({'stacklevel': 2}, **kwargs)
        self.logger.debug(message, *args, **kwargs)

    def info(self, message, *args, **kwargs):
        message = self.__wrap_message_with_uuid(message)
        kwargs = dict({'stacklevel': 2}, **kwargs)
        self.logger.info(message, *args, **kwargs)

    def exception(self, message, *args, **kwargs):
        message = self.__wrap_message_with_uuid(message)
        kwargs = dict({'exc_info': self.exc_info, 'stacklevel': 2}, **kwargs)
        self.logger.exception(message, *args, **kwargs)

    def error(self, message, *args, **kwargs):
        message = self.__wrap_message_with_uuid(message)
        kwargs = dict({'stacklevel': 2}, **kwargs)
        self.logger.error(message, *args, **kwargs)

    def warning(self, message, *args, **kwargs):
        message = self.__wrap_message_with_uuid(message)
        kwargs = dict({'stacklevel': 2}, **kwargs)
        self.logger.warning(message, *args, **kwargs)

    def fatal(self, message, *args, **kwargs):
        message = self.__wrap_message_with_uuid(message)
        kwargs = dict({'stacklevel': 2}, *kwargs)
        self.logger.critical(message, *args, **kwargs)

    def critical(self, message, *args, **kwargs):
        message = self.__wrap_message_with_uuid(message)
        kwargs = dict({'stacklevel': 2}, *kwargs)
        self.logger.critical(message, *args, **kwargs)

    def __wrap_message_with_uuid(self, message):
        message = str(message).replace('|', '').replace('\r', ' ').replace('\n', ' ')
        if self.message_uuid:
            message = f"{self.message_uuid} | {message}".strip()
        return message

    def __init_syslog_handler(self):
        """ 设置syslog日志
        """
        enable = LOGGER_ENABLE_SYSLOG_ENV_NAME
        if enable == False:
            return
        host = LOGGER_SYSLOG_HOST_ENV_NAME
        port = int(LOGGER_SYSLOG_PORT_ENV_NAME)
        facility = LOGGER_SYSLOG_FACILITY_ENV_NAME
        category = LOGGER_CATEGORY_ENV_NAME
        if not category:
            self.__create_syslog_handler(host, port, facility, logging.INFO)
        else:
            cs = category.split(",")
            for c in cs:
                level = self.__name_to_level(c)
                self.__create_syslog_handler(host, port, facility, level)

    def __name_to_level(self, name):
        return logging._nameToLevel[name]

    def __create_syslog_handler(self, host, port, facility, level):
        handler = SysLogHandler(address=(host, port), facility=SysLogHandler.facility_names[facility])
        handler.setFormatter(self.formatter)
        handler.setLevel(level)
        self._logger.addHandler(handler)

    def __init_console_handler(self):
        """ 设置终端日志
        """
        enable = LOGGER_ENABLE_CONSOLE_ENV_NAME
        if enable == False:
            return
        handler = logging.StreamHandler()
        handler.setFormatter(self.formatter)
        self._logger.addHandler(handler)

    def __init_file_handler(self):
        """ 设置文件日志
        """
        enable = LOGGER_ENABLE_FILE_ENV_NAME
        if enable == False:
            return
        directory = LOGGER_FILE_DIRECTORY_ENV_NAME
        file_utils.create_dir_if_not_exists(directory)
        category = LOGGER_CATEGORY_ENV_NAME
        cs = category.split(",")
        for c in cs:
            level = self.__name_to_level(c)
            filepath = file_utils.join_path_filename(directory, c.lower())
            self.__create_file_handler(filepath, level)

    def __create_file_handler(self, filepath, level):
        handler = logging.FileHandler(filepath)
        handler.setLevel(level)
        handler.setFormatter(self.formatter)
        self._logger.addHandler(handler)


# 初始化单例日志实例
logger = Logger()



from .http_client import http_client
from .logger import logger
from .singleton import SingletonMetaThreadSafe
from .sys_env import get_env

from .db.postgres_vector import PostgresVector
from .db.postgres_vector_async import PostgresVectorAsync


from ..util.logger import logger
from .config_template import MODEL_SETTINGS, TOKEN_SETTINGS, BASE_URL, API_KEY
import json


class ConfigManager:
    
    def __init__(self, config_file=None):
        """
        __init__ 初始化配置管理器。

        Args:
            config_file: 配置文件路径。如果不提供，则使用默认配置。
        """
        self.model_settings = []
        self.token_settings = {}
        self.load_config(config_file)   

    def load_config(self, config_file=None):
        """加载配置文件中的配置到内存。"""
        if config_file:
            # 从文件中加载配置
            with open(config_file, 'r') as f:
                config = json.load(f)
            self.model_settings = config.get('MODEL_SETTINGS', MODEL_SETTINGS)
            self.token_settings = config.get('TOKEN_SETTINGS', TOKEN_SETTINGS)
            self.base_url = config.get('BASE_URL', BASE_URL if BASE_URL else "https://www.lazygpt.cn/api")
            self.apikey = config.get('API_KEY', API_KEY if API_KEY else "XXX")
        else:
            # 使用默认配置
            self.model_settings = MODEL_SETTINGS
            self.token_settings = TOKEN_SETTINGS
            self.base_url = BASE_URL if BASE_URL else "https://www.lazygpt.cn/api"
            self.apikey = API_KEY if API_KEY else "XXX"


    def get_model_config(self, model_name):
        """根据模型名称获取配置。"""
        for config in self.model_settings:
            if config['model'] == model_name:
                return config
        return None
    
    def list_models_config(self):
        """将类中的属性转换为JSON输出。"""
        # 获取所有实例属性
        config_dict = {key: value for key, value in self.__dict__.items() if not key.startswith('_')}
        return json.dumps(config_dict, indent=4)

    def set_model_config(self, model_name, max_tokens, token_setting="gpt-3.5-turbo"):
        """设置或更新模型的最大令牌数，可选更新令牌设置名称。如果模型不存在，则添加新的配置项。"""
        found = False
        for config in self.model_settings:
            if config['model'] == model_name:
                config['max_tokens'] = max_tokens
                if token_setting is not None:
                    config['token_setting'] = token_setting
                found = True
                break
        if not found:
            new_config = {
                "model": model_name, 
                "max_tokens": max_tokens
            }
            if token_setting:
                new_config['token_setting'] = token_setting
            self.model_settings.append(new_config)


config_manager = ConfigManager()


# -*- coding: utf-8 -*-

import pytz
import time
from datetime import datetime

# import maya


class DateTime(object):

    # 时间戳常量(秒级)
    ONE_SECOND = 1
    ONE_MINUTE = 60 * ONE_SECOND
    ONE_HOUR = 60 * ONE_MINUTE
    ONE_DAY = 24 * ONE_HOUR
    ONE_WEEK = 7 * ONE_DAY
    MILLISECONDS_IN_ONE_SECOND = 1000
    NANOSECONDS_IN_ONE_SECOND = 1000000000

    # 时区
    TIMEZONE_UTC = 'utc'
    TIMEZONE_BEIJING = 'Etc/GMT-8'

    @classmethod
    def timestamp_second(cls):
        """秒级时间戳(十位)"""
        return int(time.time())

    @classmethod
    def timestamp_millisecond(cls):
        """毫秒级时间戳(十三位)"""
        return int(time.time() * cls.MILLISECONDS_IN_ONE_SECOND)

    @classmethod
    def timestamp_nanoseconds(cls):
        """纳秒级时间戳(十九位)"""
        return int(time.time() * cls.NANOSECONDS_IN_ONE_SECOND)

    @property
    def seconds(self):
        return int(self.timestamp / self.NANOSECONDS_IN_ONE_SECOND)

    @property
    def milliseconds(self):
        return int(self.timestamp / 1000000)

    @property
    def nanoseconds(self):
        return self.timestamp

    @property
    def datetime(self):
        utc = pytz.timezone(self.TIMEZONE_UTC)
        return datetime.fromtimestamp(self.seconds, tz=utc).replace(tzinfo=utc).astimezone(self.timezone)

    @property
    def year(self):
        return self.datetime.year

    @property
    def month(self):
        return self.datetime.month

    @property
    def day(self):
        return self.datetime.day

    @property
    def hour(self):
        return self.datetime.hour

    @property
    def minute(self):
        return self.datetime.minute

    @property
    def second(self):
        return self.datetime.second

    @property
    def weekday(self):
        return self.datetime.weekday()

    def __init__(self,
                 seconds=None,
                 milliseconds=None,
                 nanoseconds=None,
                 date_string=None,
                 str_format='%Y-%m-%d %H:%M:%S',
                 timezone=TIMEZONE_BEIJING):
        self.timezone = pytz.timezone(timezone)
        self.reset_time(seconds=seconds,
                        milliseconds=milliseconds,
                        nanoseconds=nanoseconds,
                        date_string=date_string,
                        str_format=str_format)

    def reset_time(self, seconds=None, milliseconds=None, nanoseconds=None, date_string=None, str_format='%Y-%m-%d %H:%M:%S'):
        # 初始化时间
        if seconds is not None:
            self.timestamp = seconds * self.NANOSECONDS_IN_ONE_SECOND
        elif milliseconds:
            self.timestamp = milliseconds * 1000000
        elif nanoseconds:
            self.timestamp = nanoseconds
        elif date_string:
            date = datetime.strptime(date_string, str_format).replace(tzinfo=self.timezone).astimezone(self.timezone)
            self.timestamp = date.timestamp() * self.NANOSECONDS_IN_ONE_SECOND
        else:
            self.timestamp = self.timestamp_nanoseconds()

    def set_timezone(self, timezone):
        self.timezone = pytz.timezone(timezone)
        return self

    def get_date_string(self, format='%Y-%m-%d %H:%M:%S'):
        return self.datetime.strftime(format)

    def add_years(self, years):
        new_year = self.datetime.year + years
        self.timestamp = int(self.datetime.replace(year=new_year).timestamp() * self.NANOSECONDS_IN_ONE_SECOND)
        return self

    def add_months(self, months):
        new_month = self.datetime.month + months
        new_year = self.datetime.year
        if new_month <= 0:
            new_year -= int(-new_month / 12)
            new_year -= 1
            new_month = 12 - (-new_month % 12)
        else:
            new_year += int(new_month / 12)
            new_month = new_month % 12
            if new_month == 0:
                new_month = 12
        self.timestamp = int(self.datetime.replace(year=new_year, month=new_month).timestamp() * self.NANOSECONDS_IN_ONE_SECOND)
        return self

    def add_weeks(self, weeks):
        self.timestamp += weeks * 7 * self.ONE_DAY * self.NANOSECONDS_IN_ONE_SECOND
        return self

    def add_days(self, days):
        self.timestamp += days * self.ONE_DAY * self.NANOSECONDS_IN_ONE_SECOND
        return self

    def add_hours(self, hours):
        self.timestamp += hours * self.ONE_HOUR * self.NANOSECONDS_IN_ONE_SECOND
        return self

    def add_minutes(self, minutes):
        self.timestamp += minutes * self.ONE_MINUTE * self.NANOSECONDS_IN_ONE_SECOND
        return self

    def add_seconds(self, seconds):
        self.timestamp += seconds * self.ONE_SECOND * self.NANOSECONDS_IN_ONE_SECOND
        return self

    def is_weekday(self):
        return self.datetime.weekday() < 5

    def is_weekend(self):
        return not self.is_weekday()

    def reset_minute(self):
        date_format = '%Y-%m-%d %H:%M'
        date_str = self.get_date_string(format=date_format)
        self.reset_time(date_string=date_str, str_format=date_format)
        return self

    def reset_hour(self):
        date_format = '%Y-%m-%d %H'
        date_str = self.get_date_string(format=date_format)
        self.reset_time(date_string=date_str, str_format=date_format)
        return self

    def reset_day(self):
        date_format = '%Y-%m-%d'
        date_str = self.get_date_string(format=date_format)
        self.reset_time(date_string=date_str, str_format=date_format)
        return self

    def reset_week(self):
        self.reset_day()
        self.add_days(-self.datetime.weekday())
        return self

    def reset_month(self):
        date_format = '%Y-%m'
        date_str = self.get_date_string(format=date_format)
        self.reset_time(date_string=date_str, str_format=date_format)
        return self

    def reset_year(self):
        date_format = '%Y'
        date_str = self.get_date_string(format=date_format)
        self.reset_time(date_string=date_str, str_format=date_format)
        return self



from ..config.config_manager import config_manager
from ..util.logger import logger

import tiktoken
import json


class Session(object):
    """这里主要维护的是大语言模型的上下轮会话"""

    def __init__(self, session_id, system_prompt):
        self.session_id = session_id
        self.messages = []
        self.system_prompt = system_prompt
        
    def to_json(self):
        return json.dumps(self.__dict__)

    @classmethod
    def from_json(cls, json_str, model):
        session_dict = json.loads(json_str)
        session = cls(session_dict['session_id'], system_prompt=session_dict.get('system_prompt', None), model=model)
        session.messages = session_dict.get('messages', [])
        return session

    # 重置会话
    def reset(self):
        system_item = {"role": "system", "content": self.system_prompt}
        self.messages = [system_item]

    def set_system_prompt(self, system_prompt):
        self.system_prompt = system_prompt
        self.reset()

    def add_query(self, query):
        user_item = {"role": "user", "content": query}
        self.messages.append(user_item)

    def add_reply(self, reply):
        assistant_item = {"role": "assistant", "content": reply}
        self.messages.append(assistant_item)

    
    def discard_exceeding(self):
        raise NotImplementedError

    def calc_tokens(self):
        raise NotImplementedError
    
    # refer to https://github.com/openai/openai-cookbook/blob/main/examples/How_to_count_tokens_with_tiktoken.ipynb
    def num_tokens_from_messages(messages, model):
        """Returns the number of tokens used by a list of messages based on dynamically loaded model settings."""
        model_config = config_manager.get_model_config(model)
        
        if not model_config:
            logger.debug(f"Model configuration not found for: {model}")
            return 0

        # Get the token settings for the specified model
        token_setting = config_manager.token_settings.get(model_config['token_setting'], {})
        tokens_per_message = token_setting.get('tokens_per_message', 3)  # Default value if not specified
        tokens_per_name = token_setting.get('tokens_per_name', 1)       # Default value if not specified

        # Get encoding using tiktoken
        try:
            encoding = tiktoken.encoding_for_model(model_config['token_setting'])
        except KeyError:
            logger.debug("Warning: model not found. Using cl100k_base encoding.")
            encoding = tiktoken.get_encoding("cl100k_base")

        num_tokens = 0
        for message in messages:
            num_tokens += tokens_per_message
            for key, value in message.items():
                if key in ['role', 'content']:
                    num_tokens += len(encoding.encode(value))
                if key == "name":
                    num_tokens += tokens_per_name

        num_tokens += 3  # Assuming every reply is primed with "assistant"
        return num_tokens
    
    def num_tokens_by_character(messages):
        """Returns the number of tokens used by a list of messages."""
        tokens = 0
        for msg in messages:
            tokens += len(msg["content"])
        return tokens




"""用户进程内部通讯的的公共变量"""
from ...util.singleton import SingletonMetaThreadSafe as SingletonMetaclass
from ...util.logger import logger
import json

"""
    e.g.  [
        {"role": "system", "content": "You are a helpful assistant."},
        {"role": "user", "content": "Who won the world series in 2020?"},
        {"role": "assistant", "content": "The Los Angeles Dodgers won the World Series in 2020."},
        {"role": "user", "content": "Where was it played?"}
    ]
"""


class GlobalStorage(metaclass=SingletonMetaclass):
    """
    如果没有传入redis客户端，则用全局变量管理会话
    """

    def __init__(self):
        self.global_map = {}
    
    def set(self, key, data, expired=24 * 60 * 60) -> bool:
        if not key:
            raise Exception("找不到key")
        self.global_map[key] = json.dumps(data)
    
    def get(self, key) -> dict:
        if not key:
            raise Exception("找不到key")
        data = self.global_map.get(key)
        return json.loads(data) if data else None
    
    def delete(self, key) -> bool:
        if not key:
            raise Exception("找不到key")
        if key in self.global_map:
            del self.global_map[key]
        return True



global_storage = GlobalStorage()


from ...prompt_base_builder import BaseBuilder
from ...prompt_template import KB_REFERENCE_TEMPLATE
from ...util import logger
from ...session_manager import SessionManagerAsync


class PromptBuilderAsync(BaseBuilder):

    @property
    def session(self):
        return self._session

    def __init__(self, session_id, session_manager: SessionManagerAsync):
            self.session_id = session_id
            self.session_manager = session_manager
            self._session = None
            self._tasks = []
            super().__init__(session_id, session_manager)

    def set_system(self, system_template, **kwargs):
        """设置系统提示词

        Args:
            system_prompt (str): 提示词模板，
            kwargs： 提示词变量渲染参数
        """
        self._tasks.append((self._async_set_system, system_template, kwargs))
        return self

    async def _async_set_system(self, system_template, **kwargs):
        self._set_template(system_template)
        self._render_template(**kwargs)
        logger.info(f"set_system content: {self.content}")
        self._session = await self.session_manager.build_session(self.session_id, self.content)
        logger.info(f"set_system session: {self._session.messages}")

    def set_prompt(self, prompt_template, **kwargs):
        self._tasks.append((self._async_set_prompt, prompt_template, kwargs))
        return self

    async def _async_set_prompt(self, prompt_template, **kwargs):
        self._set_template(prompt_template)
        self._render_template(**kwargs)
        logger.info(f"set_prompt content: {self.content}")
        self._session = await self.session_manager.session_query(self.session_id, self.content)
        logger.info(f"set_prompt session: {self._session.messages}")

    def set_kb_reference(self, **kwargs):
        self._tasks.append((self._async_set_kb_reference, KB_REFERENCE_TEMPLATE, kwargs))
        return self

    async def _async_set_kb_reference(self, kb_reference_template, **kwargs):
        self._set_template(kb_reference_template)
        self._render_template(**kwargs)
        logger.info(f"set_kb_reference content: {self.content}")
        self._session = await self.session_manager.session_query(self.session_id, self.content)
        logger.info(f"set_kb_reference session: {self._session.messages}")

    async def build(self):
        """等待所有异步任务完成"""
        for task, template, kwargs in self._tasks:
            await task(template, **kwargs)
        return self

    




# -*- coding: utf-8 -*-

from threading import Lock


class SingletonMetaThreadSafe(type):
    """
    线程安全的单例类的metaclass
    >>> class BusinessClass(metaclass=SingletonMetaThreadSafe):
    >>>     pass
    """
    _instances = {}
    _lock: Lock = Lock()

    def __call__(cls, *args, **kwargs):
        with cls._lock:
            if cls not in cls._instances:
                instance = super().__call__(*args, **kwargs)
                cls._instances[cls] = instance
        return cls._instances[cls]


class SingletonMetaNoThreadSafe(type):
    """ 非线程安全的单例metaclass
    >>> class BusinessClass(metaclass=SingletonMetaNoThreadSafe):
    >>>     pass
    """
    _instances = {}

    def __call__(cls, *args, **kargs):
        if cls not in cls._instances:
            instance = super().__call__(*args, **kargs)
            cls._instances[cls] = instance
        return cls._instances[cls]



# -*- coding: utf-8 -*-
from ...util.singleton import SingletonMetaThreadSafe
import aioredis
import json


class RedisStorageAsync(metaclass=SingletonMetaThreadSafe):

    def __init__(self, url) -> None:
        self._redis_async = aioredis.from_url(
                url,
                encoding="utf-8",
                decode_responses=True
            )
    
    async def acquire_lock(self, lock_name, lock_timeout=60):
        if lock_timeout == -1:
            # 当不希望键过期时，省略 EX 参数
            result = await self._redis_async.set(lock_name, 1, nx=True)
        else:
            # 确保过期时间是一个正整数
            result = await self._redis_async.set(lock_name, 1, ex=max(1, int(lock_timeout)), nx=True)
        return result is not None


    async def release_lock(self, lock_name):
        """释放锁。"""
        await self._redis_async.delete(lock_name)
    
    async def set(self, key, data, expired=None) -> bool:
        if not key:
            raise Exception("找不到key")
        if expired is None:
            # 没有过期时间，使用set命令
            return await self._redis_async.set(key, json.dumps(data))
        else:
            # 有过期时间，使用setex命令
            return await self._redis_async.setex(key, expired, json.dumps(data))

    async def get(self, key):
        if not key:
            raise Exception("找不到key")
        data = await self._redis_async.get(key)
        return json.loads(data) if data else None

    async def check(self, key, expired=24 * 60 * 60) -> bool:
        if not key:
            raise Exception("找不到key")
        return await self._redis_async.expire(key, expired)

    async def delete(self, key) -> None:
        if not key:
            raise Exception("找不到key")
        await self._redis_async.delete(key)

    async def close(self):
        if self._redis_async:
            self._redis_async.close()
            await self._redis_async.wait_closed()
    
    async def enqueue_message(self, queue_name, message) -> None:
        """异步将消息添加到队列中"""
        await self._redis_async.lpush(queue_name, json.dumps(message))

    async def dequeue_message(self, queue_name, timeout=0) -> dict:
        """异步从队列中移除并返回一条消息"""
        message = await self._redis_async.brpop(queue_name, timeout=timeout)
        if message:
            return json.loads(message[1])
        return None

    async def get_queue_length(self, queue_name):
        """异步获取队列长度"""
        return await self._redis_async.llen(queue_name)
    



from ..prompt.builder_async import PromptBuilderAsync

from ...util.id_generator import generate_common_id


class BaseBuilder(PromptBuilderAsync):

    @staticmethod
    def validate_chat_args(args, model):
        """
        校验传入的参数是否有效，并返回有效的参数。
        :param args: 传入的参数字典
        :return: 过滤后的有效参数字典
        """
        valid_params = {
            'frequency_penalty': lambda x: isinstance(x, (int, float)) and -2.0 <= x <= 2.0,
            'logit_bias': lambda x: isinstance(x, dict) and all(isinstance(k, str) and isinstance(v, (int, float)) and -100 <= v <= 100 for k, v in x.items()),
            'logprobs': lambda x: x in [True, False, None],
            'top_logprobs': lambda x: isinstance(x, int) and 0 <= x <= 20,
            'max_tokens': lambda x: isinstance(x, int) and x > 0,
            'n': lambda x: isinstance(x, int) and x > 0,
            'presence_penalty': lambda x: isinstance(x, (int, float)) and -2.0 <= x <= 2.0,
            'response_format': lambda x: isinstance(x, dict) and x.get("type") == "json_object" and model in ["gpt-4o", "gpt-4-turbo" , "gpt-3.5-turbo"],
            'seed': lambda x: isinstance(x, int) and x >= 0,
            'stop': lambda x: isinstance(x, (str, list)) and (isinstance(x, str) or all(isinstance(i, str) for i in x)),
            'stream_options': lambda x: isinstance(x, dict) or x is None,
            'temperature': lambda x: isinstance(x, (int, float)) and 0 <= x <= 2,
            'top_p': lambda x: isinstance(x, (int, float)) and 0 <= x <= 1,
            'tools': lambda x: isinstance(x, list) and all(isinstance(i, dict) for i in x),
            'tool_choice': lambda x: isinstance(x, (str, dict)) or x is None,
            'user': lambda x: isinstance(x, str)
        }

        validated_args = {}
        for param, validator in valid_params.items():
            if param in args and validator(args[param]):
                validated_args[param] = args[param]
            elif param in args:
                raise ValueError(f"Invalid value for parameter '{param}': {args[param]}")
        
        return validated_args

    @property
    def current_plugin(self):
        return self.__current_plugin

    def __init__(self, session_id, session_manager):
        """_summary_

        Args:
            session_id (_type_): 会话ID
            session_storage (_type_, optional): _description_. 会话存储器，默认为None，则为全局变量管理会话，目前还可以配置redis存储.
            sessioncls (_type_, optional): _description_. token管理器，目前只有GPT一种所以可以不传.
        """
        if not session_id:
            session_id = generate_common_id()
        super().__init__(session_id, session_manager)
        self.__current_plugin = None
    
    def set_plugin(self, plugin, **args):
        """动态添加属性到LLM实例。

        Args:
            attr_name (str): 属性名，例如 'kb'
            attr_value (object): 属性值，可以是任何对象，例如 KnowledgeBase 实例
        """
        self.__current_plugin = plugin(**args)


    


    




from setuptools import setup, find_packages

setup(
    name="gpts_builder",
    version="0.1.19",
    author="yancyyu",
    author_email="yancyyu.ok@gmail.com",
    description="A Python library for quickly building applications with large language models.",
    long_description=open('README.md', 'r', encoding='utf-8').read(),
    long_description_content_type='text/markdown',
    license="MIT",
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'aioredis>=1.3,<3',
        'anyio==4.3.0',
        'async-timeout==4.0.3',
        'backoff==2.2.1',
        'certifi==2024.2.2',
        'charset-normalizer==3.3.2',
        'httpx==0.27.0',
        'redis==5.0.4',
        'regex==2024.4.28',
        'requests==2.31.0',
        'sniffio==1.3.1',
        'tiktoken==0.6.0',
        'typing_extensions==4.11.0',
        'urllib3==2.2.1',
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    url="https://github.com/yancyuu/gpts_builder.git",  # Optional
    keywords="GPT, large language models, AI, application builder",
    zip_safe=False
)



from ..llm.llm_builder import LLM
from ...util.db.postgres_vector import PostgresVector
from ...util.db.schema import EmbbedingSchame, DatasetSchema, KB_TABLE_NAME, ANSWER_TABLE_NAME, QUESTION_TABLE_NAME
from ...util.id_generator import generate_common_id
from ...util.logger import logger
import json, re
from scipy.spatial.distance import cosine


class DatasetBuilder:        

    def __init__(self, db_driver: PostgresVector):
        """初始化知识库插件构建器。

        Args:
            connection_params (dict): 数据库连接参数。
        """
        super().__init__()
        self.db_driver = db_driver
        self.db = self.db_driver.db
        self.answer_schema = EmbbedingSchame(ANSWER_TABLE_NAME)
        self.question_schema = EmbbedingSchame(QUESTION_TABLE_NAME)
        self.dataset_schema = DatasetSchema()

    def create_dataset(self, name, creator="gpt_builder"):
        """
        向知识库添加一个新的条目，包括创建者和创建时间。

        Args:
            name (str): 知识库条目的名称。
            creator (str): 条目的创建者。

        Returns:
            int: 新创建的条目的 ID。
        """
        with self.db.cursor() as cursor:
            # 插入新条目并返回生成的 ID
            cursor.execute(f"""
            INSERT INTO {KB_TABLE_NAME} 
            ({self.dataset_schema.name}, {self.dataset_schema.creator}, {self.dataset_schema.created_at}) 
            VALUES (%s, %s, CURRENT_TIMESTAMP) RETURNING {self.dataset_schema.id}
            """, (name, creator))
            
            # 获取并返回新插入条目的 ID
            new_id = cursor.fetchone()[0]
            self.db.commit()  # 确保提交数据库事务以保存插入
            return new_id
            
    def get_dataset(self, filters ,creator="gpt_builder"):
        """
        根据指定的过滤条件获取知识库条目。

        参数:
            creator (str, optional): 创建者的名称。默认为 "gpt_builder"。
            filters (dict): 过滤条件的键值对。

        返回:
            int or None: 如果找到条目，则返回其 ID；否则返回 None。
        """
        # 构建过滤条件的 SQL 语句
        filter_conditions = " AND ".join([f"{key} = %s" for key in filters.keys()])
        filter_values = list(filters.values())

        # 添加 creator 过滤条件
        filter_conditions += f" AND {self.dataset_schema.creator} = %s"
        filter_values.append(creator)

        query = f"""
        SELECT * FROM {KB_TABLE_NAME}
        WHERE {filter_conditions}
        """

        with self.db.cursor() as cursor:
            cursor.execute(query, filter_values)
            result = cursor.fetchall()
            if result:
                return result  # 返回找到的条目
            else:
                return None  # 如果没有找到，返回 None

    def create_datas(self, id, index_text, output_list):
        """向数据库添加一个答案和相关的问题列表。"""
        if not id:
            logger.info("未成功生产知识库实例")
            return
        try:
            # 开启事务
            cursor = self.db.cursor()
            
            # 插入索引并获取ID
            sql_query = self.db_driver.get_insert_into_str(ANSWER_TABLE_NAME, [self.answer_schema.text, self.answer_schema.vector, self.answer_schema.kb_id]) + f" RETURNING {self.answer_schema.id}"
            index_vector = self.generate_vector(index_text).get("embedding")
            if not index_vector:
                logger.info(f"{index_text} 未成功向量化")
                return
            cursor.execute(sql_query, (index_text, index_vector, id))
            index_id = cursor.fetchone()[0]
            logger.info(f"index_id {index_id}")
            
            # 批量插入问题
            question_entries = []
            for question_text in output_list:
                question_vector = self.generate_vector(question_text).get("embedding")
                if question_vector:
                    question_entries.append((index_id, question_text, question_vector))
                else:
                    logger.info(f"{question_text} 未成功向量化")
            
            if question_entries:
                sql_query = self.db_driver.get_insert_into_str(QUESTION_TABLE_NAME, [self.answer_schema.id, self.question_schema.text, self.question_schema.vector]) + f" RETURNING {self.question_schema.id}"
                cursor.executemany(sql_query, question_entries)
            
            self.db.commit()  # 提交事务
            cursor.close()  # 关闭游标
            return True
        except Exception as e:
            logger.error(f"数据库操作失败: {e}")
            self.db.rollback()  # 回滚事务
            cursor.close()  # 关闭游标
            return False
            


    def generate_vector(self, text):
        """生成文本的向量表示，使用spaCy库。

        Args:
            text (str): 输入文本。

        Returns:
            blob: 向量的BLOB表示。
        """
        response = LLM.embedding(text)
        if not response:
            return {}
        return response.get("data",{}).get("data", [])[0] if response.get("data") else {}

    def query_similarity(self, text, dataset_ids=[], similarity=-1, search_all=False):
        """
        计算输入文本与数据库中内容的余弦相似度，并返回包括索引信息和相关内容信息的数据结构。

        Args:
            dataset_ids (list): 知识库ID列表。
            text (str): 查询的文本。
            similarity (float): 用于过滤的相似度阈值。
            search_all (bool): 是否计算加权平均相似度。

        Returns:
            dict: 以知识库ID为键，每个键包含相关内容的列表。
        """
        cursor = self.db.cursor()
        
        if not dataset_ids:
            logger.info("未成功获取知识库实例")
            return None

        input_vector = self.generate_vector(text).get("embedding")
        if not input_vector:
            logger.info("输入文本未成功向量化")
            return None

        kb_content_map = {}

        for kb_id in dataset_ids:
            try:
                # 查询每个kb_id对应的答案和问题数据
                sql = f"""
                    SELECT a.{self.answer_schema.vector}, a.{self.answer_schema.text}, a.{self.answer_schema.id},
                        q.{self.question_schema.vector}, q.{self.question_schema.text}
                    FROM {ANSWER_TABLE_NAME} AS a
                    JOIN {QUESTION_TABLE_NAME} AS q ON a.{self.answer_schema.id} = q.{self.answer_schema.id}
                    WHERE a.{self.answer_schema.kb_id} = %s"""
                logger.info(f"[gpt_builder] SQL:{sql}")
                cursor.execute(sql, (kb_id,))
                data_list = cursor.fetchall()

                content_list = []
                for answer_vector, answer_text, answer_id, question_vector, question_text in data_list:
                    answer_vector_data = json.loads(answer_vector)
                    question_vector_data = json.loads(question_vector)
                    answer_similarity = 1 - cosine(input_vector, answer_vector_data)
                    question_similarity = 1 - cosine(input_vector, question_vector_data)
                    total_similarity = (answer_similarity + question_similarity) / 2 if search_all else question_similarity

                    if total_similarity < similarity:
                        continue

                    content_list.append({
                        "index_id": answer_id,
                        "related_questions": question_text,
                        "answer_text": answer_text,
                        "similarity": total_similarity,
                        "search_all": search_all
                    })

                if content_list:
                    content_list.sort(key=lambda x: x['similarity'], reverse=True)
                    kb_content_map[kb_id] = content_list

            except Exception as e:
                logger.error(f"处理知识库 {kb_id} 时出错: {e}")

        return kb_content_map

    def query_regex(self, regex, dataset_ids=[]):
        """
        根据正则表达式查询数据库中的内容。

        Args:
            regex (str): 查询的正则表达式。
            dataset_ids (list): 知识库ID列表。

        Returns:
            dict: 以知识库ID为键，每个键包含匹配的索引ID和对应的索引文本及相关内容列表的字典。
        """
        cursor = self.db.cursor()
        if not dataset_ids:
            logger.info("未成功获取知识库实例")
            return None

        # 检查正则表达式是否有效
        try:
            re.compile(regex)
        except re.error as e:
            logger.error(f"无效的正则表达式: {regex}, 错误: {e}")
            return None

        # 最外层字典，以kb_id为键
        kb_content_map = {}

        for kb_id in dataset_ids:
            try:
                # 使用连表查询每个kb_id对应的答案和问题数据
                sql = f"""
                    SELECT a.{self.answer_schema.id} AS answer_id, a.{self.answer_schema.text} AS answer_text, q.{self.question_schema.id} AS question_id, q.{self.question_schema.text} AS question_text
                    FROM {ANSWER_TABLE_NAME} AS a
                    JOIN {QUESTION_TABLE_NAME} AS q ON a.{self.answer_schema.id} = q.{self.answer_schema.id}
                    WHERE a.{self.answer_schema.kb_id} = %s AND q.{self.question_schema.text} ~ %s"""
                logger.info(f"[gpt_builder] SQL:{sql} param {kb_id, regex}")
                cursor.execute(sql, (kb_id, regex))
                data_list = cursor.fetchall()

                for answer_id, answer_text, question_id, question_text in data_list:
                    if kb_id not in kb_content_map:
                        kb_content_map[kb_id] = []

                    # 添加到列表
                    kb_content_map[kb_id].append({
                        "answer_id": answer_id,
                        "answer_text": answer_text,
                        "question_id": question_id,
                        "related_questions": question_text
                    })

            except Exception as e:
                logger.error(f"处理知识库 {kb_id} 时出错: {e}")

        return kb_content_map
        


from .llm_base_builder import BaseBuilder
from ...config.config_manager import config_manager
from ...session_manager.session_manager import ChatGPTSession
from ...session_manager.session_manager_async import SessionManagerAsync
from ...session_manager.storage.redis_storage_async import RedisStorageAsync

from ...util.http_client import http_client
from ...util.logger import logger

import json


class LLMAsync(BaseBuilder):

    @property
    def current_plugin(self):
        return self.__current_plugin

    def __init__(self, model, session_storage: RedisStorageAsync, session_id=None, sessioncls = ChatGPTSession):
        """_summary_

        Args:
            session_id (_type_): 会话ID，如果为空，会自动生成一个
            session_storage (_type_, optional): _description_. 会话存储器，异步模式必须配置redis存储.
            sessioncls (_type_, optional): _description_. token管理器，目前只有GPT一种所以可以不传.
        """
        if not config_manager.get_model_config(model):
            raise Exception("没有此模型配置")
        self.session_manager = SessionManagerAsync(sessioncls, session_storage, model)
        super().__init__(session_id=session_id, session_manager=self.session_manager)
        self.__current_plugin = None


    @staticmethod
    async def embedding(input):
        """向OpenAI发送请求，获取文本的embedding"""
        headers = {
            'Authorization': f"Bearer {config_manager.apikey}",
            'Content-Type': 'application/json'
        }
        
        # 准备请求数据
        payload = {
            "input": input,
            "model": "text-embedding-ada-002"
        }
        url = config_manager.base_url + "/v1/embeddings"

        curl_command = BaseBuilder.generate_curl_command(url, payload, headers) 
        logger.info(f"[gpts-builder] {curl_command}")

        response = await http_client.post_async(url=url, json=payload, headers=headers)
        # 记录响应
        logger.info(f"[gpts-builder] Received embedding response: {response}")

        return response
    
    async def chat_completions(self, ** args) -> ChatGPTSession:
        """大模型chat请求，非流式返回
        """
        try:
            # 校验传入的参数
            valid_args = self.validate_chat_args(args, self.session_manager.model)
        except ValueError as e:
            raise Exception(f"Invalid argument: {e}")
        # 1. 准备请求头
        headers = {
            'Authorization': f"Bearer {config_manager.apikey}",
            'Content-Type': 'application/json'
        }
        # 2. 准备请求数据
        payload = {
            "messages": self.session.messages,
            "model": self.session.model
            }
        if args:
            payload.update(valid_args)
        url = config_manager.base_url + "/v1/chat/completions"

        curl_command = BaseBuilder.generate_curl_command(url, payload, headers) 
        logger.info(f"[gpts-builder] {curl_command}")

        return await http_client.post_async(url=url, json=payload, headers=headers, timeout=60, max_retries=3)
    
    async def chat_completions_stream(self, **args):
        """大模型chat请求，流式返回一个异步生成器"""
        try:
            # 校验传入的参数
            valid_args = self.validate_chat_args(args, self.session_manager.model)
        except ValueError as e:
            raise Exception(f"[gpts-builder] Invalid argument: {e}")
        
        # 准备请求头
        headers = {
            'Authorization': f"Bearer {config_manager.apikey}",
            'Content-Type': 'application/json'
        }
        
        # 准备请求数据
        payload = {
            "messages": self.session.messages,
            "model": self.session.model,
            "stream": True
        }
        if args:
            payload.update(valid_args)

        url = config_manager.base_url + "/v1/chat/completions"        
        curl_command = BaseBuilder.generate_curl_command(url, payload, headers) 
        logger.info(f"[gpts-builder] {curl_command}")

        async for line in http_client.post_stream_async(url=url, json=payload, headers=headers, timeout=60):
            yield line 
                        
    async def clear_session(self):
        """清除会话"""
        try:
            await self.session_manager.clear_session(self.session_id)
        except Exception as e:
            raise Exception(f"Failed to clear session: {e}")


    




import requests
import httpx
import backoff

class HTTPClient:
    
    def __init__(self):
        # 创建一个持久的 AsyncClient 实例
        self.async_client = httpx.AsyncClient()
    
    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        # 确保在类实例不再使用时关闭 client
        await self.async_client.aclose()

    def post(self, url, json=None, headers=None, timeout=60, max_retries=3):
        """
        执行同步的 HTTP POST请求，并支持自定义超时和重试次数。
        
        参数:
            url (str): 请求的 URL。
            json (dict, optional): 这是要发送的数据。默认为 None。
            timeout (int, optional): 请求超时时间（秒）。默认为 60 秒。
            max_retries (int, optional): 失败时的最大重试次数。默认为 3。
        
        返回:
            dict: 从服务器返回的 JSON 数据。
        
        异常:
            HTTPError: 如果响应状态码表示错误。
        """
        @backoff.on_exception(backoff.expo,
                              (requests.exceptions.RequestException, requests.exceptions.HTTPError, requests.exceptions.Timeout),
                              max_tries=max_retries)
        def request_with_backoff():
            response = requests.post(url, json=json, timeout=timeout, headers=headers)
            response.raise_for_status()
            return response.json()

        return request_with_backoff()
    
    def post_stream(self, url, json=None, headers=None, timeout=60, max_retries=3):
        """
        执行同步的 HTTP POST 请求，并支持流式响应、自定义超时和重试次数。
        
        参数:
            url (str): 请求的 URL。
            json (dict, optional): 这是要发送的数据。默认为 None。
            timeout (int, optional): 请求超时时间（秒）。默认为 60 秒。
            headers (dict, optional): 请求的头部。默认为 None。
            max_retries (int, optional): 失败时的最大重试次数。默认为 3。
        
        返回:
            Generator[str, None, None]: 流式响应内容的生成器。
        
        异常:
            HTTPError: 如果响应状态码表示错误。
        """
        
        @backoff.on_exception(backoff.expo,
                              (requests.exceptions.RequestException, requests.exceptions.HTTPError, requests.exceptions.Timeout),
                              max_tries=max_retries)
        def request_with_backoff():
            response = requests.post(url, json=json, headers=headers, timeout=timeout, stream=True)
            response.raise_for_status()
            for chunk in response.iter_content(chunk_size=1024, decode_unicode=True):
                yield chunk
        
        return request_with_backoff()

    def get(self, url, params=None, headers=None, timeout=60, max_retries=3):
        """
        执行同步的 HTTP GET请求，并支持自定义超时和重试次数。
        
        参数:
            url (str): 请求的 URL。
            params (dict, optional): 要发送的数据。默认为 None。
            timeout (int, optional): 请求超时时间（秒）。默认为 60 秒。
            max_retries (int, optional): 失败时的最大重试次数。默认为 3。
        
        返回:
            dict: 从服务器返回的 JSON 数据。
        
        异常:
            HTTPError: 如果响应状态码表示错误。
        """
        @backoff.on_exception(backoff.expo,
                              (requests.exceptions.RequestException, requests.exceptions.HTTPError, requests.exceptions.Timeout),
                              max_tries=max_retries)
        def request_with_backoff():
            response = requests.get(url, params=params, timeout=timeout, headers=headers)
            response.raise_for_status()
            return response.json()

        return request_with_backoff()

    async def post_async(self, url, json=None, headers=None, timeout=60, max_retries=3):
        """
        执行异步的 POST 请求，并支持自定义超时和重试次数。
        
        参数:
            url (str): 请求的 URL。
            json (dict, optional): 这是要发送的数据。默认为 None。
            timeout (int, optional): 请求超时时间（秒）。默认为 60 秒。
            max_retries (int, optional): 失败时的最大重试次数。默认为 3。
        
        返回:
            dict: 从服务器返回的 JSON 数据。
        
        异常:
            ValueError: 如果提供了不支持的 HTTP 方法。
        """
        @backoff.on_exception(backoff.expo,
                            (httpx.TimeoutException, httpx.RequestError, httpx.HTTPStatusError, httpx.HTTPError),
                              max_tries=max_retries)
        async def request_with_backoff():
            response = await self.async_client.post(url, json=json, timeout=timeout, headers=headers)
            response.raise_for_status()
            return response.json()
        
        return await request_with_backoff()
        
    async def post_stream_async(self, url, json=None, headers=None, timeout=60, max_retries=3):
        """
        执行异步的 POST 请求，并支持自定义超时和重试次数。
        
        参数:
            url (str): 请求的 URL。
            json (dict, optional): 这是要发送的数据。默认为 None。
            timeout (int, optional): 请求超时时间（秒）。默认为 60 秒。
            max_retries (int, optional): 失败时的最大重试次数。默认为 3。
        
        返回:
            dict: 从服务器返回的 JSON 数据。
        
        异常:
            ValueError: 如果提供了不支持的 HTTP 方法。
        """
        @backoff.on_exception(backoff.expo,
                            (httpx.TimeoutException, httpx.RequestError, httpx.HTTPStatusError, httpx.HTTPError),
                            max_tries=max_retries)
        async def request_with_backoff():
            async with httpx.AsyncClient() as client:
                # Properly using the async with statement with client.stream
                async with client.stream("POST", url=url, headers=headers, json=json, timeout=timeout) as response:
                    response.raise_for_status()  # Ensure the response status is OK
                    async for chunk in response.aiter_text(chunk_size=1024):
                        yield chunk
        # 返回生成器
        async for item in request_with_backoff():
            yield item
    
    async def get_async(self, url, params=None, headers=None, timeout=60, max_retries=3):
        """
        执行异步的 HTTP GET 或 POST 请求，并支持自定义超时和重试次数。
        
        参数:
            url (str): 请求的 URL。
            params (dict, optional): 这是要发送的数据。默认为 None。
            timeout (int, optional): 请求超时时间（秒）。默认为 60 秒。
            max_retries (int, optional): 失败时的最大重试次数。默认为 3。
        
        返回:
            dict: 从服务器返回的 JSON 数据。
        
        异常:
            HTTPError: 如果响应状态码表示错误。
            ValueError: 如果提供了不支持的 HTTP 方法。
        """
        @backoff.on_exception(backoff.expo,
                            (httpx.TimeoutException, httpx.RequestError, httpx.HTTPStatusError, httpx.HTTPError),
                              max_tries=max_retries)
        async def request_with_backoff():
            response = await self.async_client.get(url, params=params, timeout=timeout, headers=headers)
            response.raise_for_status()
            return response.json()

        return await request_with_backoff()
    
    

http_client = HTTPClient()


KB_REFERENCE_TEMPLATE = \
                """使用代码块中的内容作为你的辅助知识，并且根据这些知识按下面的要求回答问题:
                    ```{reference}```

                    回答要求：
                    - 仔细理解辅助知识和问题直接的关联，如果辅助知识中的内容和问题毫不相关，则拒绝回答问题。
                    - 如果答案中有引用辅助知识，请确保引用内容和辅助知识中一致，不要胡编乱造。
                    - 使用 Markdown 语法优化回答格式，按照逻辑分段回答。
                    - 使用与问题相同的语言回答。

                    问题是:{question}
                """